{- Heist.hs (c) 2009 Claude Heiland-Allen <claude@mathr.co.uk> -}

{-# LANGUAGE FlexibleContexts, GADTs #-}

module Dot where

import qualified Prelude as P
import Prelude (Bool, Float, Int, Integer, String, (++), (.),
  concat, cycle, drop, error, filter, foldr, length, map, pi, putStrLn,
  repeat, show, snd, take, undefined, unlines, zip, zipWith, zipWith3)
import Data.List ((\\), intercalate)
import Data.Maybe (fromJust, isJust)
import qualified Data.Set as S
import qualified Data.Map as M
import Types.Data.Num

import Heist

-- diagram generation

type Id = Int
type Diagram = ([(Id, Int, Int, String)], [(Id, Int)], [(Id, Int)], [((Id, Int),(Id, Int))])

toDiagram :: [Id] -> Bd (i,o) -> (Diagram, [Id], M.Map String Id, M.Map String (S.Set Id))

toDiagram n (Rec   a b) = let ((a'ns, a'is, a'os, a'es), n', m, mm) = toDiagram n  a
                              ((b'ns, b'is, b'os, b'es), n'', m', mm') = toDiagram n' b
                              zn = length (zip a'os b'is)
                              n''' = drop zn n''
                              z'ns = take zn (map (\z -> (z,1,1,"z^-1")) n'')
                              z'es = concat (zipWith3 (\a b z -> [(a,(z,1)),((z,1),b)]) a'os b'is n'')
                          in ((a'ns ++ b'ns ++ z'ns,
                               drop (length b'os) a'is,
                               a'os,
                               a'es ++ b'es ++ (zip b'os a'is) ++ z'es
                              ), n''', m `M.union` m', M.unionWith S.union mm mm')

toDiagram n (Par   a b) = let ((a'ns, a'is, a'os, a'es), n', m, mm) = toDiagram n  a
                              ((b'ns, b'is, b'os, b'es), n'', m', mm') = toDiagram n' b
                          in ((a'ns ++ b'ns,
                               a'is ++ b'is,
                               a'os ++ b'os,
                               a'es ++ b'es
                              ), n'', m `M.union` m', M.unionWith S.union mm mm')

toDiagram n (Seq   a b) = let ((a'ns, a'is, a'os, a'es), n', m, mm) = toDiagram n  a
                              ((b'ns, b'is, b'os, b'es), n'', m', mm') = toDiagram n' b
                              mi = P.min (length a'os) (length b'is)
                          in ((a'ns ++ b'ns,
                               a'is ++ drop mi b'is,
                               drop mi a'os ++ b'os,
                               a'es ++ b'es ++ (zip a'os b'is)
                              ), n'', m `M.union` m', M.unionWith S.union mm mm')

toDiagram n (Split a b) = let ((a'ns, a'is, a'os, a'es), n', m, mm) = toDiagram n  a
                              ((b'ns, b'is, b'os, b'es), n'', m', mm') = toDiagram n' b
                          in ((a'ns ++ b'ns,
                               a'is,
                               b'os,
                               a'es ++ b'es ++ (zip (cycle a'os) b'is)
                              ), n'', m `M.union` m', M.unionWith S.union mm mm')

toDiagram n (Merge a b) = let ((a'ns, a'is, a'os, a'es), n', m, mm) = toDiagram n  a
                              ((b'ns, b'is, b'os, b'es), n'', m', mm') = toDiagram n' b
                          in ((a'ns ++ b'ns,
                               a'is,
                               b'os,
                               a'es ++ b'es ++ (zip a'os (cycle b'is))
                              ), n'', m `M.union` m', M.unionWith S.union mm mm')

toDiagram n (Prim s t)  = let i = fromIntegerT (lengthT (fstT t))
                              o = fromIntegerT (lengthT (sndT t))
                              is = zip (take i n) (repeat 1)
                              os = zip (take o (drop i n)) (repeat 1)
                              (me:n') = drop o (drop i n)
                          in (([(me, i, o, s)],
                               is,
                               os,
                               zipWith (\x p -> (x, (me, p))) is [1..i] ++
                               zipWith (\x p -> ((me, p), x)) os [1..o]
                              ), n', M.empty, M.empty)

toDiagram n (Get s v)   = let (o:me:n') = n
                          in (([(me, 0, 1, v)],
                               [],
                               [(o,1)],
                               [((me,1),(o,1))]
                              ), n', M.empty, M.insert v (S.insert me S.empty) M.empty)

toDiagram n (Put s v)   = let (i:me:n') = n
                          in (([(me, 1, 1 {- output port for later... -} , v)],
                               [(i,1)],
                               [],
                               [((i,1),(me,1))]
                              ), n', M.insert v me M.empty, M.empty)

-- add the top level inputs and outputs, and compact put -> get interactions
addInOut :: Int -> Int -> (Diagram, [Id], M.Map String Id, M.Map String (S.Set Id)) -> (Diagram, S.Set Id)
addInOut ins outs ((ns, is, os, es), (i:o:ks), ps, gss)
  = ((inNode : outNode : filter isnGN ns, is, os, filter isnGE (fs ++ varEdges)), puts)
  where
    inNode   = (i, 0, ins, "IN")
    outNode  = (o, outs, 0, "OUT")
    inEdges  = [ ((i,k),t) | (k, t) <- zip [1..] is ]
    outEdges = [ (f,(o,k)) | (k, f) <- zip [1..] os ]
    fs       = inEdges ++ outEdges ++ es
    varEdges = [ ((p,1),t)
               | (n,p) <- M.toList ps
               , let jgs = M.lookup n gss
               , isJust jgs
               , g <- S.toList (fromJust jgs)
               , (f,t) <- fs
               , f P.== (g,1)
               ]
    isnGN (n,_,_,_) = P.not (isGet n)
    isnGE ((n,_),_) = P.not (isGet n)
    isGet n = n `S.member` gets
    gets = (S.unions (map snd (M.toList gss)))
    puts = S.fromList (map snd (M.toList ps))

-- transitively compact identity wires together
noWires :: (Diagram, S.Set Id) -> Diagram
noWires ((ns, is, os, es), ps) = (filter (\(n,_,_,_) -> P.not (n `S.member` ps)) ns, is, os, go es)
  where
    go [] = []
    go (e@((f,o),(t,i)):xs) =
      if f `S.member` nodes P.&& t `S.member` nodes
       then e : go xs                                       -- f node, t node
       else if f `S.member` nodes
             then go (map (replaceFrom (t,i) (f,o)) xs)     -- f node, t wire
             else if t `S.member` nodes
                   then go (map (replaceTo (f,o) (t,i)) xs) -- f wire, t node
                   else go (map (replaceTo (f,o) (t,i)) xs) -- f wire, t wire
    replaceFrom replace with (src, dest) = (if src P.== replace then with else src, dest)
    replaceTo   replace with (src, dest) = (src, if dest P.== replace then with else dest)
    nodes = S.fromList (map (\(n,_,_,_) -> n) ns) S.\\ ps

-- convert a block diagram to an (optimized) GraphViz .dot format
toDot :: (NaturalT (Length i), NaturalT (Length o)) => Bd (i,o) -> String
toDot bd =
  let (ns, is, os, es) = noWires (addInOut ins outs (toDiagram [0..] bd))
      ins = fromIntegerT (lengthT (fstT (bdT bd)))
      outs = fromIntegerT (lengthT (sndT (bdT bd)))
  in unlines
    [ "digraph G {"
    , "graph [ rankdir = \"LR\" ];"
    , "node [ shape=\"record\" ];" 
    , unlines (map namedNode ns)
    , unlines (map edge es)
    , "}"
    ]
  where
    name n = "n" ++ (show n)
    nameP (n,p) = "n" ++ (show n) ++ ":" ++ (show p)
    namedNode (n, i, o, s) = name n ++ toLabel s i o
    edge ((e,o),(f,i)) = name e ++ ":" ++ oport o ++ "->"
                      ++ name f ++ ":" ++ iport i ++ ";"
    toLabel s i o = " [ " ++
       "label=\"{{" ++ intercalate "|" (map (wrap . iport) [1..i]) ++ "}|"
       ++ s ++ "|{" ++ intercalate "|" (map (wrap . oport) [1..o]) ++ "}}\" ];"
    wrap s = "<" ++ s ++ ">"
    iport i = "i" ++ (show i)
    oport o = "o" ++ (show o)

