{- Heist.hs (c) 2009 Claude Heiland-Allen <claude@mathr.co.uk> -}

{-# LANGUAGE
        ExistentialQuantification,
        FlexibleContexts,
        FlexibleInstances,
        GADTs,
        MultiParamTypeClasses,
        ScopedTypeVariables,
        TypeFamilies,
        TypeOperators,
        TypeSynonymInstances,
        UndecidableInstances #-}

module Heist where

-- Basic requirements
import qualified Prelude as P
import Prelude (Bool, Float, Int, Integer, String, (++), (.),
  concat, cycle, drop, error, filter, foldr, length, map, pi, putStrLn,
  repeat, show, snd, take, undefined, unlines, zip, zipWith, zipWith3)
import Data.List ((\\), intercalate)
import Data.Maybe (fromJust, isJust)
import qualified Data.Set as S
import qualified Data.Map as M

-- Advanced requirements
import Types.Data.Num hiding (Mod, (:.))
import Types.Data.Num.Decimal
import Types.Data.Ord
import Types.Data.Bool

-- Core types
type instance Compare Bool Bool = EQ
type instance Compare Int Int = EQ
type instance Compare Float Float = EQ
type instance Compare String String = EQ

-- Type-level numbers mod algorithm
type family Mod' x y b
type instance Mod' x y False = x
type instance Mod' x y True  = Mod' (x :-: y) y (((x :-: y) :>=: y) :&&: (y :>: D0))
type family Mod x y
type instance Mod x y = Mod' x y ((x :>=: y) :&&: (y :>: D0))
type family MultipleOf x y
type instance MultipleOf x y = (Mod x y) :==: D0

-- Type-level lists
data Nil' x xs = Nil'
type Nil = Nil' () ()
data x :. xs = x :. xs
infixr 9 :.
type instance Compare Nil Nil = EQ
type instance Compare (x :. xs) (y :. ys) = If (x :==: y) (Compare xs ys) (Compare x y)

-- Type-level list length
type family Length l
type instance Length Nil = D0
type instance Length (x :. xs) = D1 :+: Length xs
lengthT :: l -> Length l
lengthT _ = error "lengthT"

-- Type-level list take
type family Take' n l n_gt_z
type instance Take' n l False = Nil
type instance Take' n (x :. xs) True = x :. Take' (n :-: D1) xs ((n :-: D1) :>: D0)
type family Take n l
type instance Take n l = Take' n l (n :>: D0)

-- Type-level list drop
type family Drop' n l n_gt_z
type instance Drop' n l False = l
type instance Drop' n (x :. xs) True = Drop' (n :-: D1) xs ((n :-: D1) :>: D0)
type family Drop n l
type instance Drop n l = Drop' n l (n :>: D0)

-- Type-level list comparison
type family Equals x y
type instance Equals Nil Nil = True
type instance Equals (x :. xs) (y :. ys) = (x :==: y) :&&: (xs `Equals` ys)

-- Type-level list append
type family Append x y
type instance Append Nil ys = ys
type instance Append (x :. xs) ys = x :. Append xs ys

-- Type-level list cycle fused with take
type family TakeCycle' n xs l_gt_z
type instance TakeCycle' n xs False = Nil
type instance TakeCycle' n xs True = Append xs (TakeCycle' (n :-: Length xs) xs
                                                          ((n :-: Length xs) :>: D0))
type family TakeCycle n xs
type instance TakeCycle n xs = TakeCycle' n xs (n :>: D0)

-- Type-level list replicate
type family Replicate' n t n_gt_z
type instance Replicate' n t False = Nil
type instance Replicate' n t True = t :. (Replicate' (n :-: D1) t ((n :-: D1) :>: D0))
type family Replicate n t
type instance Replicate n t = Replicate' n t (n :>: D0)

-- Type-level list foldr
type family FoldR (f :: * -> * -> *) e xs
type instance FoldR f e Nil = e
type instance FoldR f e (x :. xs) = f x (FoldR f e xs)

-- Type-level pairs
type family Fst x
type instance Fst (a,b) = a
type family Snd x
type instance Snd (a,b) = b
fstT :: (i,o) -> i
fstT _ = error "fstT"
sndT :: (i,o) -> o
sndT _ = error "sndT"

bdT :: Bd t -> t
bdT _ = error "bdT"

-- Recursive composition of block diagrams
type family IsRecCompat x y
type instance IsRecCompat (xi, xo) (yi, yo) = (Take (Length yi) xo `Equals` yi)
                                         :&&: (Take (Length yo) xi `Equals` yo)
type family RecShape x y
type instance RecShape (xi, xo) (yi, yo) = (Drop (Length yo) xi, xo)

-- Parallel composition of block diagrams
type family IsParCompat x y
type instance IsParCompat (xi, xo) (yi, yo) = True
type family ParShape x y
type instance ParShape (xi, xo) (yi, yo) = (xi `Append` yi, xo `Append` yo)

-- Sequential composition of block diagrams
type family IsSeqCompat x y
type instance IsSeqCompat (xi, xo) (yi, yo) = Take (Min (Length xo) (Length yi)) xo
                                     `Equals` Take (Min (Length xo) (Length yi)) yi
type family SeqShape x y
type instance SeqShape (xi, xo) (yi, yo) = (xi `Append` (Drop (Min (Length xo) (Length yi)) yi),
                                            yo `Append` (Drop (Min (Length xo) (Length yi)) xo))

-- Split composition of block diagrams
type family IsSplitCompat x y
type instance IsSplitCompat (xi, xo) (yi, yo) = IsPositive (Length xo)
                                           :&&: MultipleOf (Length yi) (Length xo)
                                           :&&: TakeCycle (Length yi) xo `Equals` yi
type family SplitShape x y
type instance SplitShape (xi, xo) (yi, yo) = (xi, yo)

-- Merge composition of block diagrams
type family IsMergeCompat x y
type instance IsMergeCompat (xi, xo) (yi, yo) = IsPositive (Length yi)
                                           :&&: MultipleOf (Length xo) (Length yi)
                                           :&&: TakeCycle (Length xo) yi `Equals` xo
                                           :&&: AllMergable yi
type family MergeShape x y
type instance MergeShape (xi, xo) (yi, yo) = (xi, yo)

-- Check if types supports merging
type family AllMergable x
type instance AllMergable Nil = True
type instance AllMergable (x :. xs) = IsMergable x :&&: AllMergable xs
type family IsMergable t
type instance IsMergable Int = True
type instance IsMergable Float = True

-- How to merge
class MergeOp a where
  mergeOp :: [a] -> a
instance MergeOp Int where
  mergeOp = foldr (P.+) 0
instance MergeOp Float where
  mergeOp = foldr (P.+) 0

-- Generalized Abstract block diagrams
data Bd t where

  Rec     :: (i ~ (ai, ao), j ~ (bi, bo), k ~ (ci, co),
              NaturalT (Length ai), NaturalT (Length ao),
              NaturalT (Length bi), NaturalT (Length bo),
              NaturalT (Length ci), NaturalT (Length co),
              IsRecCompat   i j ~ True, RecShape   i j ~ k) => Bd i -> Bd j -> Bd k

  Par     :: (i ~ (ai, ao), j ~ (bi, bo), k ~ (ci, co),
              NaturalT (Length ai), NaturalT (Length ao),
              NaturalT (Length bi), NaturalT (Length bo),
              NaturalT (Length ci), NaturalT (Length co),
              IsParCompat   i j ~ True, ParShape   i j ~ k) => Bd i -> Bd j -> Bd k

  Seq     :: (i ~ (ai, ao), j ~ (bi, bo), j ~ (ci, co),
              NaturalT (Length ai), NaturalT (Length ao),
              NaturalT (Length bi), NaturalT (Length bo),
              NaturalT (Length ci), NaturalT (Length co),
              IsSeqCompat   i j ~ True, SeqShape   i j ~ k) => Bd i -> Bd j -> Bd k

  Split   :: (i ~ (ai, ao), j ~ (bi, bo), j ~ (ci, co),
              NaturalT (Length ai), NaturalT (Length ao),
              NaturalT (Length bi), NaturalT (Length bo),
              NaturalT (Length ci), NaturalT (Length co),
              IsSplitCompat i j ~ True, SplitShape i j ~ k) => Bd i -> Bd j -> Bd k

  Merge   :: (i ~ (ai, ao), j ~ (bi, bo), j ~ (ci, co),
              NaturalT (Length ai), NaturalT (Length ao),
              NaturalT (Length bi), NaturalT (Length bo),
              NaturalT (Length ci), NaturalT (Length co),
              IsMergeCompat i j ~ True, MergeShape i j ~ k) => Bd i -> Bd j -> Bd k

  Prim    :: (k ~ (ci, co),
              NaturalT (Length ci), NaturalT (Length co))   => String  -> k -> Bd k

  Put     ::                                                   s -> String -> Bd (s :. Nil, Nil)

  Get     ::                                                   s -> String -> Bd (Nil, s :. Nil)

-- Fixities as in FAuSt
infixl 6 *, /
infixl 5 +, -, `max`, `min`
infixl 4 `Rec`
infixl 3 `Par`
infixl 2 `Seq`
infixl 1 `Split`
infixl 1 `Merge`

-- n-ary parallel composition

type Empty = Bd (Nil, Nil)
empty :: Empty
empty = Prim "empty" (error "empty")

class Parallel' b n j where
  type Parallel'ShapeIn b n j
  type Parallel'ShapeOut b n j
  parallel' :: b -> n -> (Integer -> Bd j) -> Bd (Parallel'ShapeIn b n j, Parallel'ShapeOut b n j)

instance Parallel' True D0 j where
  type Parallel'ShapeIn  True D0 j = Nil
  type Parallel'ShapeOut True D0 j = Nil
  parallel' _ _ _ = empty

instance ( IntegerT (n :-: D1),
           NaturalT (Length (Parallel'ShapeIn  ((n :-: D1) :==: D0) (n :-: D1) (bi, bo))),
           NaturalT (Length (Parallel'ShapeOut ((n :-: D1) :==: D0) (n :-: D1) (bi, bo))),
           NaturalT (Length bi),
           NaturalT (Length bo),
           NaturalT (Length (Append (Parallel'ShapeIn  ((n :-: D1) :==: D0) (n :-: D1) (bi, bo)) bi)),
           NaturalT (Length (Append (Parallel'ShapeOut ((n :-: D1) :==: D0) (n :-: D1) (bi, bo)) bo)),
           Parallel' ((n :-: D1) :==: D0) (n :-: D1) (bi, bo)
         ) => Parallel' False n (bi,bo) where
  type Parallel'ShapeIn  False n (bi,bo) = Fst (ParShape (Parallel'ShapeIn ((n:-:D1) :==: D0) (n :-: D1) (bi,bo), Parallel'ShapeOut ((n:-:D1) :==: D0) (n :-: D1) (bi,bo)) (bi,bo))
  type Parallel'ShapeOut False n (bi,bo) = Snd (ParShape (Parallel'ShapeIn ((n:-:D1) :==: D0) (n :-: D1) (bi,bo), Parallel'ShapeOut ((n:-:D1) :==: D0) (n :-: D1) (bi,bo)) (bi,bo))
  parallel' _ n f = parallel' ((n `subT` d1) `eqT` d0) (n `subT` d1) f `Par` f (fromIntegerT (n`subT` d1))

class Parallel n j where
  parallel :: Parallel' (n :==: D0) n j => n -> (Integer -> Bd j) -> Bd (Parallel'ShapeIn (n :==: D0) n j, Parallel'ShapeOut (n :==: D0) n j)
  parallel n f = parallel' (n `eqT` d0) n f

instance NaturalT n => Parallel n j where -- default


-- common primitive shapes
type ToI = (Nil, Int :. Nil)
type FTo = (Float :. Nil, Nil)
type ToF = (Nil, Float :. Nil)
type FToF = (Float :. Nil, Float :. Nil)
type FFToF = (Float :. Float :. Nil, Float :. Nil)
fftof = undefined :: FFToF

-- primitives
const x = Prim (show x) (error (show x) :: ToF)
mul = Prim "*" (error "*" :: FFToF)
div = Prim "/" (error "/" :: FFToF)
add = Prim "+" (error "+" :: FFToF)
sub = Prim "-" (error "-" :: FFToF)
umax = Prim "max" (error "max" :: FFToF)
umin = Prim "min" (error "min" :: FFToF)
(*) x y = x `Par` y `Seq` mul
(/) x y = x `Par` y `Seq` div
(+) x y = x `Par` y `Seq` add
(-) x y = x `Par` y `Seq` sub
max x y = x `Par` y `Seq` umax
min x y = x `Par` y `Seq` umin
usin = Prim "sin" (error "sin" :: FToF)
ucos = Prim "cos" (error "cos" :: FToF)
sin x = x `Seq` usin
cos x = x `Seq` ucos
wire x = Put f x `Seq` Get f x

-- common wire types
i = undefined :: Int
f = undefined :: Float

